
package org.nla.cobol;

import java.io.LineNumberReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Map;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.Iterator;
import java.util.StringTokenizer;

/**
 * Parser of COBOL CopyBooks  
 * @author NLA
 */
public class CobolCopyBookParser
{		

	/////////////////////////////////////////////////////////////////
	///		Constants
	/////////////////////////////////////////////////////////////////

	/** position of comments in copybook */
	private static int INT_POS_COMMENTS = 7;
	
	/** separator of keywords and lexemes */
	private static String STR_SEPARATOR = " ";
	
	/** dot element */
	private static String STR_DOT = ".";
	
	/** PIC keyword */
	private static String STR_PIC = " pic ";
	
	/** PICTURE keyword */
	private static String STR_PICTURE = " picture ";
	
	/** OCCURS keyword */
	private static String STR_OCCURS = " occurs ";
	
	/** COMP-3 keyword */
	private static String STR_COMP3 = " comp-3";
	
	/** COMPUTATIONAL-3 keyword */
	private static String STR_COMPU3 = " computational-3";

	/** USAGE keyword */
	private static String STR_USAGE = " usage ";
	
	/** BINARY keyword */
	private static String STR_BINARY = " binary";
	
	/** PACKED-DECIMAL keyword */
	private static String STR_PCKDDEC = " packed-decimal";
	
	/** VALUE keyword */
	private static String STR_VALUE = " value ";
	
	

	/////////////////////////////////////////////////////////////////
	///		Members
	/////////////////////////////////////////////////////////////////
	
	/** internal reader, aimed to read copybook content */
	private LineNumberReader _reader;
	
	/** mapping between line numbers and line contents */
	private Map<Integer, String> _mapLines;
	
	/** internal iterator on line numbers */
	private Iterator<Integer> _itLineNumbers;
	
	/** internal copybook */
	private CobolCopyBook _copy; 


	/////////////////////////////////////////////////////////////////
	///		Constructors
	/////////////////////////////////////////////////////////////////
	
	/**
	 * Constructs a with the copybook path file. This constructor is not to be called by package users.
	 * @param	path	path to the copybook file
	 * @throws	IOException, if a problem is encountered with the file
	 */
	private CobolCopyBookParser(String path) throws IOException {
		_reader = new LineNumberReader( new FileReader( path ) );
		_mapLines = new LinkedHashMap<Integer, String>();
	}
	
	
	/////////////////////////////////////////////////////////////////
	///		Static methods
	/////////////////////////////////////////////////////////////////
	
	/**
	 * Returns a copybook created with the file designed by <i>path</i>
	 * @param	path	path to the copybook file
	 * @throws	IOException, if a problem is encountered with the file
	 */
	public static CobolCopyBook buildCopyBook( String path )	throws IOException  {
		CobolCopyBookParser ccbp = new CobolCopyBookParser( path );
		return ( ccbp.buildCopyBook() );
	}

	
	/////////////////////////////////////////////////////////////////
	///		Methods
	/////////////////////////////////////////////////////////////////
	
	/**
	 * Returns a copybook created with the file created by the constructor.
	 * @throws	IOException, if a problem is encountered with the file
	 */
	private CobolCopyBook buildCopyBook() throws IOException {
		readFileContent();
		_copy = new CobolCopyBook();
		_itLineNumbers = _mapLines.keySet().iterator();
		ArrayList<CobolElement> lstElements = buildRawElementsFromFile();
		_copy.buildCopyBookFromRawElements( lstElements );

		return _copy;
	}

	/**
	 * Reads content of the file, line by line. Each line is added to the internal map, identified by its number.
	 * Commented and empty lines are left aside. 
	 * @throws	IOException, if a problem is encountered with the file
	 */
	private void readFileContent()  throws IOException {
		for ( String strLine = _reader.readLine(); strLine != null; strLine = _reader.readLine() ) {
			if ( 
				strLine.trim().length() != 0 && 
				strLine.charAt(INT_POS_COMMENTS-1) != '*'
				) {
				_mapLines.put( _reader.getLineNumber(), strLine ); 
			}
		}
	}

	/**
	 * Builds and return cobol elements from the lines previously read. These elements don't have any knowledge of each others, 
	 * they are neither grouped or hierarchied, hence the term "raw"
	 * @return	list of raw cobol elements
	 */
	private ArrayList<CobolElement> buildRawElementsFromFile()
	{	
		ArrayList<CobolElement> lstElements = new ArrayList<CobolElement>();
		//_itLineNumbers = _mapLines.keySet().iterator();
		
		while ( _itLineNumbers.hasNext() ) {
			lstElements.addAll( buildElementFromLine() );
		}
		
		return lstElements;
	}
	
	/**
	 * Builds and return a single cobol element from a line previously read.
	 * @return	a single cobol element  
	 */
	private ArrayList<CobolElement> buildElementFromLine()
	{
		ArrayList<CobolElement> lstElement = new ArrayList<CobolElement>();
		
		int intNbLine = _itLineNumbers.next();
		String strLine = _mapLines.get(intNbLine);
		
		// remove first characters up to comment position
		strLine = strLine.substring( INT_POS_COMMENTS );
		strLine = strLine.trim();
		
		if ( !strLine.endsWith(STR_DOT) ) {
			throw ( new RuntimeException("CobolCopyBookParser::buildElementFromLine() Line ["+intNbLine+"] doesn't end with '.'") );
		}
		// remove . at the end
		strLine = strLine.substring( 0, strLine.length() - 1 );
		
		StringTokenizer sttkSpace = new StringTokenizer( strLine, STR_SEPARATOR, false );
		
		if ( isLineAGroup(strLine) ) {
			CobolVariableGroup cvg = new CobolVariableGroup();
			cvg.setOurParentOfUsAll(_copy);
			String strLevel = sttkSpace.nextToken(); 			// group level
			String strName = sttkSpace.nextToken(); 			// group name
		
			if ( isOccursed(strLine) ) {
				String ss = sttkSpace.nextToken(); 							// OCCURS keyword
				String strOccursCount = sttkSpace.nextToken(); 	// occurs count
				
				cvg.setOccursed(true);
				cvg.setOccursCount( Integer.parseInt(strOccursCount) );
			}
			
			cvg.setName( strName );
			cvg.setLevelAsString( strLevel );
			cvg.setOurParentOfUsAll( _copy );
			lstElement.add( cvg );
			lstElement.addAll( buildRawElementsFromFile() );
		}
		else {
			CobolVariable cv = new CobolVariable();
			cv.setOurParentOfUsAll(_copy);
			
			String strLevel = sttkSpace.nextToken(); 			// variable level
			String strName = sttkSpace.nextToken(); 			// variable name
			sttkSpace.nextToken(); 								// PIC or PICTURE keyword
			String strType = sttkSpace.nextToken(); 			// type and length
			
			if ( isComp3(strLine) ) {
				sttkSpace.nextToken(); 							// one of COMP-3 synonyms
				cv.setComp3(true);
			}

			if ( hasValue(strLine) ) {
				sttkSpace.nextToken(); 							// VALUE keyword
				sttkSpace.nextToken();							// actual value
			}
			
			if ( isOccursed(strLine) ) {
				sttkSpace.nextToken(); 							// OCCURS keyword
				String strOccursCount = sttkSpace.nextToken(); 	// occurs count
				
				cv.setOccursed(true);
				cv.setOccursCount( Integer.parseInt(strOccursCount) );
			}
			cv.setName( strName );
			cv.setLevelAsString( strLevel );
			
			// parse type
			Map mapTypeInfos = buildTypeInfos(strType);
			String strCobolType = (String)mapTypeInfos.get("INNERTYPE");
			Integer intIntSize = (Integer)mapTypeInfos.get("INTSIZE");
			Integer intDecSize = (Integer)mapTypeInfos.get("DECSIZE");
			Boolean isDecimal = (Boolean)mapTypeInfos.get("ISDECIMAL");
			Boolean isSigned = (Boolean)mapTypeInfos.get("ISSIGNED");
			
			cv.setCobolType( strCobolType.equals("9") ? CobolType.NUM : CobolType.CHAR );
			cv.setIntSize( intIntSize.intValue() );
			cv.setDecSize( intDecSize.intValue() );
			cv.setDecimal( isDecimal.booleanValue() );
			cv.setSigned( isSigned.booleanValue() );
			cv.setOurParentOfUsAll( _copy );
					
			lstElement.add( cv );
		}
		
		return lstElement;
	}

	/////////////////////////////////////////////////////////////////
	///		Static Methods
	/////////////////////////////////////////////////////////////////	
	
	/**
	 * Identifies and returns the characteristics of the type of an element. 
	 * These characteristics are returned in a map, and identified as such : <br>
	 * &nbsp;&nbsp;- INNERTYPE : the inner cobol type of the element X or 9 <br>
	 * &nbsp;&nbsp;- INTSIZE : the integer size of the element <br>
	 * &nbsp;&nbsp;- DECSIZE : the decimal size of the element <br>
	 * &nbsp;&nbsp;- ISDECIMAL : the decimal state of the element <br>
	 * &nbsp;&nbsp;- ISSIGNED : the signed state of the element <br><br>
	 * <b>WARNING : This method can't yet read types where the integer or the decimal sizes 
	 * are defined with several Xs or 9s. For instance, 9999V999 isn't properly interpreted.
	 * As for now, it could instead be written 9(4)V9(3).</b> 
	 * @param 	strType		Raw type as found in a line
	 * @return	a map of all type characteristics
	 */
	private static Map buildTypeInfos( String strType ) {
	
		HashMap map = new HashMap();
		map.put("INNERTYPE", "" );
		map.put("INTSIZE", 1 );
		map.put("DECSIZE", 0 );
		map.put("ISDECIMAL", false );
		map.put("ISSIGNED", false );
		
		String strFirstLetter = strType.substring(0, 1);
		if ( strFirstLetter.equalsIgnoreCase("S") ) {
			map.put("ISSIGNED", true );
			strType = strType.substring(1);
		}
		
		String strInnerType = strType.substring(0, 1);
		map.put("INNERTYPE", strInnerType );
		
		if ( strType.length() > 1 ) {
			// V or (
			String strCharAfterInnerType = strType.substring(1, 2);
			
			// integer size determination
			if ( strCharAfterInnerType.equals("V") ) {
				map.put("INTSIZE", 1 );
			}
			else if ( strCharAfterInnerType.equals("(") ) {
				int intFirstLeftBracket = strType.indexOf( "(" );
				int intFirstRightBracket = strType.indexOf( ")" );
				map.put("INTSIZE", Integer.parseInt(strType.substring(intFirstLeftBracket+1, intFirstRightBracket)) );
			}
			
			if ( strInnerType.equals("X") ) {
				return map;
			}
		
			// decimal size determination
			int intVPosition = strType.indexOf("V");
			if ( intVPosition != -1 ) {
				int intV9Position = strType.indexOf("V9");
				if ( intV9Position != -1 ) {
					int intSecondLeftBracket = strType.lastIndexOf( "(" );
					int intSecondRightBracket = strType.lastIndexOf( ")" );
					map.put("DECSIZE", Integer.parseInt(strType.substring(intSecondLeftBracket+1, intSecondRightBracket)) );
					map.put("ISDECIMAL", true );
				}
			}
		}
		return ( map );
	}
	
	/**
	 * Tells whether a line is a variable. It returns true if the line contains
	 * either PIC or PICTURE keywords.
	 * @param 	line	A line read from the copybook file
	 * @return	true if the element represented by the line is a variable, false otherwise.
	 */
	private static boolean isLineAVariable( String line ) {
		return( line.toLowerCase().contains(STR_PIC) || line.toLowerCase().contains(STR_PICTURE) );
	}
	
	/**
	 * Tells whether a line is a group. It returns true if the line isn't a variable.
	 * @param 	line	A line read from the copybook file
	 * @return	true if the element represented by the line is a group, false otherwise.
	 */
	private static boolean isLineAGroup( String line ) {
			return( !isLineAVariable( line ) );
	}

	/**
	 * Tells whether the element in the line is occursed. It returns true if the line contains
	 * the OCCURS keywords.
	 * @param 	line	A line read from the copybook file
	 * @return	true if the element represented by the line is occursed, false otherwise.
	 */
	private static boolean isOccursed( String line ) {
		return( line.toLowerCase().contains(STR_OCCURS) );
	}

	/**
	 * Tells whether the element in the line is binary. It returns true if the line contains
	 * either COMP-3, BINARY or PACKED-DECIMAL keywords. <br><br>
	 * <b>WARNING : This method doesn't tell if USAGE COMPUTATIONAL is used. Another method 
	 * must be defined to that end</b> 
	 * @param 	line	A line read from the copybook file
	 * @return	true if the element represented by the line is binary, false otherwise.
	 */
	private static boolean isComp3( String line ) {
		return( 
				line.toLowerCase().contains(STR_COMP3)  ||
		//		line.toLowerCase().contains(STR_COMPU3) ||
				line.toLowerCase().contains(STR_BINARY) ||
				line.toLowerCase().contains(STR_PCKDDEC)
				);
	}
	
	/**
	 * Tells whether the element in the line has a value. It returns true if the line contains
	 * the VALUE keywords. <br><br>
	 * @param 	line	A line read from the copybook file
	 * @return	true if the element represented by the line has a value, false otherwise.
	 */	
	private static boolean hasValue( String strLine ) {
		return( strLine.toLowerCase().contains(STR_VALUE) );
	}
	
	/////////////////////////////////////////////////////////////////
	///		Test methods
	/////////////////////////////////////////////////////////////////
	
	public static void main(String[] args) throws IOException
	{
		
		CobolCopyBook c = buildCopyBook("copy.txt");
		CobolElement ce = c.findCobolElement("LETSASRP-MVT-NORM", true);
		
		System.out.println(c.getCount());
		
		
		/*System.out.println( c );
		
		 
		System.out.println( c );
		
		for ( CobolElement element : c.getListOfCobolElements() ) {
			System.out.println( element.getAbsoluteName() );
		}
		*/
		//CobolElement ce = c.findCobolElement("LETSASRP-UT[15]");
		//System.out.println( ce );
		
		
		//System.out.println( c.getChildCount() );
		

		
		
		//CobolVariableGroup e = (CobolVariableGroup)c.findCobolElement( "LETSASRP-TETBRG" );
		
		//System.out.println( c.getCount() );
		//System.out.println( c.getSize() );
		//System.out.println( c.toString() );
		
		//c.builOrderedVariables();
		//c.d();
		//System.out.println( c );
		
		
		
		/*
		Map mapTypeInfos = buildTypeInfos(args[0]);
		
		System.out.println( mapTypeInfos.get("INNERTYPE") );
		System.out.println( mapTypeInfos.get("INTSIZE") );
		System.out.println( mapTypeInfos.get("DECSIZE") );
		System.out.println( mapTypeInfos.get("ISDECIMAL") );
		*/
		
		
	}
	
}