
package org.nla.cobol;

import java.util.Map;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Iterator;
import java.util.ListIterator;

/**
 * Representation of a COBOL Group 
 * @author NLA
 */
public class CobolVariableGroup extends CobolElement implements Iterable<String>
{

	/////////////////////////////////////////////////////////////////
	///		Constants
	/////////////////////////////////////////////////////////////////
	
	private static final String STR_GROUP = "GROUP";
	
	
	/////////////////////////////////////////////////////////////////
	///		Members
	/////////////////////////////////////////////////////////////////	

	/** map containing all the elements of this group, identified by their names */
	protected Map<String, CobolElement> _mapElements;
	

	
	/////////////////////////////////////////////////////////////////
	///		Getters
	/////////////////////////////////////////////////////////////////	
	

	/** */

	
	public Map<String, CobolElement> getCobolElements() {
		return _mapElements;
	}
	
	/** 
	 *	Setters
	 */
	


	/////////////////////////////////////////////////////////////////
	///		Methods
	/////////////////////////////////////////////////////////////////	

	/**
	 *	Default and only constructor  
	 */
	public CobolVariableGroup() {
		super();
		_mapElements = new LinkedHashMap<String, CobolElement>();
	}


	/////////////////////////////////////////////////////////////////
	///		Methods
	/////////////////////////////////////////////////////////////////	
	
	/**
	 * Returns the number of direct child elements in this group
	 * @return	number of child elements
	 */
	public int getChildCount() {
		return _mapElements.size();
	}

	/**
	 * Always returns the String GROUP
	 * @return	String GROUP
	 */
	@Override
	public String getFinalType() {
		return STR_GROUP;
	}

	/**
	 * Returns the size of this group, that is the size of all its children multiplied by the number of its occurs 
	 * @return	size of this group
	 */
	@Override
	public int getSize() {		
		int intSize = 0;
		for ( CobolElement element : _mapElements.values() )	{
			intSize += element.getSize();
		}
		return intSize * _intOccursCount;
	}
	
	/**
	 * Returns the number of elements in this group and its children
	 * @return	 Returns the number of elements in this group
	 */
	@Override
	public int getCount()	{
		int intCount = 0;
		for ( CobolElement element : _mapElements.values() )	{
			intCount += element.getCount();
		}
		return intCount * _intOccursCount;
	}
	
	/**
	 * Looks for the element identified by elementName either in all its children or only in its direct children. 
	 * Returns the element as soon as it's found, null otherwise  
	 * @param 	elementName			name of the element to find 
	 * @param 	lookInAllChildren 	boolean indicating if the search must be done in all children or only in direct children
	 * @return	the actual element if found, null otherwise 
	 */
	public CobolElement findCobolElement( String elementName, boolean lookInAllChildren ) {
		if ( _mapElements.keySet().contains( elementName ) ) {
			return _mapElements.get( elementName );
		}
		else {
			if ( lookInAllChildren ) {
				for ( CobolElement element : _mapElements.values() ) {
					if ( element instanceof CobolVariableGroup ) {
						CobolVariableGroup cvg = (CobolVariableGroup)element;
						CobolElement foundElement = cvg.findCobolElement(elementName, lookInAllChildren);
						if ( foundElement != null ) {
							return foundElement;
						}
					}
				}
			}
		}
		return null;
	}
	
	/**
	 * Looks for the element identified by elementName in all its children. Returns the element as soon as it's found, 
	 * null otherwise. This method is equivalent to  findCobolElement(elementName, true)
	 * @param 	elementName			name of the element to find 
	 * @return	the actual element if found, null otherwise 
	 */
	public CobolElement getCobolElement( String elementName ) {
		return findCobolElement( elementName, true );
	}
	
	/**
	 * Looks for the element identified by elementName only in its direct children. 
	 * Returns the element if present, null otherwise  
	 * @param 	elementName	name of the element to find
	 * @return	the actual element if found, null otherwise 
	 */
	public CobolElement findCobolElementInDirectChildren( String elementName ) {
		return findCobolElement( elementName, false );
	}	
	
	/**
	 * Adds an element to its children. 
	 * If the level of this is higher than the level of the element to add, an unchecked runtime exception is thrown 
	 * @param 	cce		element to add
	 */
	public void addCobolElement( CobolElement cce )	{
		if ( cce.getLevel() <= _intLevel ) {
			throw( new RuntimeException( "CobolVariableGroup::addCobolElement(CopyCobolElement) level of sub-element " + cce.getName() + "("+cce.getLevelAsString()+") must be higher than element " + _strName + "("+_strLevel+")") );
		}
		_mapElements.put( cce.getAbsoluteName(), cce );
	}
	
	/**
	 * Adds the specified list of elements to its children. <br> 
	 * If the level of this is higher than the level of one of the elements to add, an unchecked runtime exception is thrown 
	 * @param 	lstElements		list of elements to add
	 */
	public void addCobolElements( List<CobolElement> lstElements )	{
		for ( CobolElement element : lstElements ) {
			addCobolElement( element );
		}
	}

	/**
	 * Returns an iterator on all its children names
	 * @return	an iterator on all its children names
	 */
	@Override
	public Iterator<String> iterator() {
		return _mapElements.keySet().iterator();
	}

	/**
	 * Organize raw elements into structured groups as represented in the original copybook.
	 * This method must be called only once, before buildGroups() and computePositions() 
	 * @param rootElement	root group of the copybook 
	 * @param itElements	iterator of the elements previously read 
	 */
	protected void buildGroups( CobolVariableGroup rootElement, ListIterator<CobolElement> itElements ) {
		while ( itElements.hasNext() ) {
			CobolElement element = itElements.next();
			
			// untested
			if ( element.getLevel() == 88 ) {
				return;
			}
			
			if ( element.getLevel() <= rootElement.getLevel() ) {
				itElements.previous();
				return;
			}			
			
			if ( element.isOccursed() ) {
				element.setOccursDepth( rootElement.getOccursDepth() + 1 );
				
			}
			else {
				element.setOccursDepth( rootElement.getOccursDepth() );
			}
			
			if ( element instanceof CobolVariableGroup ) {
				buildGroups( (CobolVariableGroup)element, itElements);
			}
			
			rootElement.addCobolElement( element );
		}
	}
	
	/**
	 * Build occursed elements once the raw elements have been injected in the copybook.
	 * This method must be called only once, between buildGroups() and computePositions()
	 */
	protected  void buildOccurs() {
		Map<String, CobolElement> mapFinalGroups = new LinkedHashMap<String, CobolElement>();
		for ( CobolElement element : _mapElements.values() ) {
			mapFinalGroups.put(element.getAbsoluteName(), element);
			
			// si element occurs on le clone n fois avec les bons indices puis on l'ajoute  la nouvelle map
			if ( element.isOccursed() ) {
				
				List<CobolElement> clonedElements = element.cloneItself(element.getOccursCount(), 1, _lstIndices);
				clonedElements.size();
				for ( CobolElement tmpElement : clonedElements ) {
					mapFinalGroups.put( tmpElement.getAbsoluteName(), tmpElement );
					
					if ( tmpElement instanceof CobolVariableGroup ) {
						CobolVariableGroup cvg = (CobolVariableGroup)tmpElement;
						cvg.buildOccurs();
					}
				}
			}
			// sinon, on ne fait que l'ajouter  la nouvelle map avec le bon indice
			else {
				element.setIndices( _lstIndices );
				mapFinalGroups.put( element.getAbsoluteName(), element );

				if ( element instanceof CobolVariableGroup ) {
					CobolVariableGroup cvg = (CobolVariableGroup)element;
					cvg.buildOccurs();
				}
			}
		}
		
		Iterator<String> itElementNames =  mapFinalGroups.keySet().iterator();
		while ( itElementNames.hasNext() ) {
			String strElementName = itElementNames.next();
			if ( strElementName.contains("?") ) {
				itElementNames.remove();
			}
			else {
				CobolElement ce = mapFinalGroups.get(strElementName);
				ce.setOccursed( false );
			}
		}
		
		_mapElements = mapFinalGroups;
	}
	
	/**
	 * Computes the positions of all cobol elements in the group, after the occursed elements have been built.
	 * This method must be called only once after buildGroups() and buildOccurs().
	 * @param	startIndex	Index from where the positions must be computed
	 */
	protected void computePositions( int startIndex ) {
		
		for ( String strAbsoluteName : _mapElements.keySet() ) {
			CobolElement element = _mapElements.get( strAbsoluteName );

			int intLastIndex = startIndex + element.getSize()-1;
			element.setStartPosition(startIndex);
			element.setEndPosition(intLastIndex);

			ourParentToUsAll.finalChildrenAsMap.put( strAbsoluteName, element );
			ourParentToUsAll.finalChildrenAsList.add( element );
			
			if ( element instanceof CobolVariableGroup ) {
				CobolVariableGroup cvg = (CobolVariableGroup)element;
				cvg.computePositions(startIndex);
			}
			
			startIndex = intLastIndex+1;
		}
	}

	/**
	 * Returns a new CobolVariableGroup cloned with this group's properties.
	 * @returns 	a clone of this  
	 */
	@Override
	public CobolElement cloneItself() {
		CobolVariableGroup cvg = new CobolVariableGroup();
		cvg.setOurParentOfUsAll( ourParentToUsAll );
		cvg.setLevel(_intLevel);
		cvg.setName(_strName);
		cvg.setOccursed(_blnOccurs);
		cvg.setOccursCount(_intOccursCount);
		cvg.setOccursDepth(_intOccursDepth);
		cvg.setOurParentOfUsAll( ourParentToUsAll );
		
		for ( CobolElement element : _mapElements.values() ) {
			cvg.addCobolElement( element.cloneItself() );
		}
		return ( cvg );
	}

	/**
	 * Returns a String representation of this group and all its children
	 * @returns 	 number of occurs count 
	 */
	@Override
	public String toString() {
		StringBuilder toReturn = new StringBuilder();
		
		toReturn.append( buildLevels(_intLevel) + _strLevel + " " + getAbsoluteName() + " " + (_blnOccurs ? "OCCURS x " + _intOccursCount : "" ) + "(" + _intStartPosition + "," + _intEndPosition + ")" +"\n" );
		toReturn.append( buildLevels(_intLevel) + "=>" + getChildCount() + " children\n");
		for ( CobolElement element : _mapElements.values() )	{
			toReturn.append( element ); 
		}

		return toReturn.toString();
	}
	
	/**
	 * Returns a String representation of this group and its direct children
	 * @returns 	 number of occurs count 
	 */
	public String toStringDirectChildren() {
		StringBuilder toReturn = new StringBuilder();
		toReturn.append( "VARGROUP=[" + _strLevel + "|" + getAbsoluteName() + (_blnOccurs ? "|OCCURS x " + _intOccursCount : "" ));
		for ( CobolElement element : _mapElements.values() )	{
			toReturn.append( "|" + element.getName() );
		}
		toReturn.append( "]" );
		return toReturn.toString();	
	}
	
	
	/////////////////////////////////////////////////////////////////
	///		Test Methods
	/////////////////////////////////////////////////////////////////
	
	
	public static void main(String[] args)
	{
		CobolVariable v1 = new CobolVariable();		
		v1.setLevelAsString("10");
		v1.setName("LETSASRP-INDBAS");
		v1.setIntSize(5);
		v1.setCobolType(CobolType.NUM);

		CobolVariable v2 = new CobolVariable();		
		v2.setLevelAsString("10");
		v2.setName("LETSASRP-LETIND");
		v2.setIntSize(1);
		v2.setCobolType(CobolType.CHAR);

		CobolVariable v3 = new CobolVariable();		
		v3.setLevelAsString("05");
		v3.setName("LETSASRP-DTEFF");
		v3.setIntSize(10);
		v3.setCobolType(CobolType.CHAR);

		CobolVariableGroup cvgIndice = new CobolVariableGroup();
		cvgIndice.setLevelAsString("05");
		cvgIndice.setName("LETSASRP-INDICE");
		cvgIndice.addCobolElement(v1);
		cvgIndice.addCobolElement(v2);
		
		CobolVariableGroup cvg01 = new CobolVariableGroup();
		cvg01.setLevelAsString("01");
		cvg01.setName("LETSASRP");
		cvg01.addCobolElement(cvgIndice);
		cvg01.addCobolElement(v3);
		
		
		// variables
		System.out.println(v1);
		System.out.println(v2);
		System.out.println(v3);
		
		// groups
		System.out.println("cvgIndice.getSize(): " + cvgIndice.getSize());
		System.out.println("cvgIndice: " + cvgIndice);
		System.out.println("cvg01.getSize(): " + cvg01.getSize());
		System.out.println("cvg01: " + cvg01);

		// level hierarchy problem
		/*
		CobolVariable v4 = new CobolVariable();		
		v4.setLevelAsString("01");
		v4.setName("LETSASRP-TESTKO");
		v4.setIntSize(10);
		v4.setCobolType(CobolType.CHAR);
		cvg01.addCobolElement(v4);*/

		// level numercity problem
		/*
		CobolVariable v4 = new CobolVariable();		
		v4.setName("LETSASRP-TESTKO");
		v4.setLevelAsString("RR");
		v4.setIntSize(10);
		v4.setCobolType(CobolType.CHAR);
		cvg01.addCobolElement(v4);*/
	}
	
}