
////////////////////////////////////////////////////////////////////////////////////////
//  Package
////////////////////////////////////////////////////////////////////////////////////////

package org.nla.utils;


////////////////////////////////////////////////////////////////////////////////////////
//  Imports
////////////////////////////////////////////////////////////////////////////////////////

import java.io.FileReader;
import java.io.BufferedReader;
import java.io.PrintStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Date;
import java.util.StringTokenizer;
import java.text.SimpleDateFormat;


////////////////////////////////////////////////////////////////////////////////////////
//  Class definition
////////////////////////////////////////////////////////////////////////////////////////

/**
 * This class allows to output messages on any output stream. This messages can be of three different types :<br>
 *   - begin 	: they are meant to indicate the entry in a method, potentially containing parameters <br>
 *   - end   	: they are meant to indicate the exit of a method, potentially containing parameters<br>
 *   - log	: they are meant to provide any information<br><br>
 * 
 * This class is mostly designed to help developers identify a process followed by a program. 
 * It may also be used in a production environment to probe for a problem, without neither altering performances 
 * nor preventing any end user to use the application. <br><br>
 *   
 * This simple class works with a configuration file called debug.config, that contains thee pieces information :<br>
 *   - whether it's actually activated (activated property, true or false)<br>
 *   - what timestamp format to use for the output (tsformat property, has to follow java.text.SimpleDateFormat pattern)<br>   
 *   - what packages and/or classes are being probed<br>
 * 
 * @see java.text.SimpleDateFormat
 */

public class Debug
{

	////////////////////////////////////////////////////////////////////////////////////////
	//  Constants
	 ////////////////////////////////////////////////////////////////////////////////////////

	private static final String STR_FILENAME = "debug.config";
	private static final String STR_EQUALS = "=";
	private static final String STR_DOT = ".";
	private static final String STR_RESOLUTION = "::";
	private static final String STR_DOUBLE_BRACKETS = "()";
	private static final String STR_CONSTRUCTOR = "<init>";
	private static final String STR_BEGIN = " begin";
	private static final String STR_BEGIN_WITH = " begin with ";
	private static final String STR_END = " end";
	private static final String STR_END_WITH = " end with ";
	private static final String STR_STD_FORMAT = "yyyy/MM/dd HH:mm:ss.SSS";
	private static final String STR_ACTIVATED_PROPERTY = "activated";
	private static final String STR_OUTFILE_PROPERTY = "outfile";
	private static final String STR_TSFORMAT_PROPERTY = "tsformat";


	////////////////////////////////////////////////////////////////////////////////////////
	//  Static members
	////////////////////////////////////////////////////////////////////////////////////////

	private static final ArrayList _logObjects = new ArrayList();

	private static boolean _blnActivated = false;

	private static SimpleDateFormat _dateFormat = new SimpleDateFormat( STR_STD_FORMAT );;

	private static String _strDateFormat;

	private static PrintStream _outStream = System.out;


	////////////////////////////////////////////////////////////////////////////////////////
	//  Static operations
	////////////////////////////////////////////////////////////////////////////////////////

	static
	{
		try
		{
			BufferedReader buf = new BufferedReader( new FileReader( STR_FILENAME ) );

			for ( String strLine = buf.readLine(); strLine != null; strLine = buf.readLine() )
			{
				// no line or line starting with #
				if ( strLine.length() == 0 || strLine.startsWith( "#" ) )
				{
					continue;
				}

				// activated property
				if ( strLine.toLowerCase().startsWith( STR_ACTIVATED_PROPERTY ) )
				{
					int intEqualsIndex = strLine.indexOf( STR_EQUALS );
					String strActivatedValue = strLine.substring( intEqualsIndex + 1 ).trim();
					_blnActivated = Boolean.parseBoolean( strActivatedValue );
					continue;
				}

				// outfile property
				if ( strLine.toLowerCase().startsWith( STR_OUTFILE_PROPERTY ) )
				{
					int intEqualsIndex = strLine.indexOf( STR_EQUALS );
					String strOutFile = strLine.substring( intEqualsIndex + 1 ).trim();
					try
					{
						_outStream = new PrintStream( new FileOutputStream( strOutFile, true ), true );
					}
					catch( IOException ioe )
					{
						System.out.println( "Debug::Debug() Problem using specified output file ; using standard output : " + ioe.getMessage() );
					}
					continue;
				}

				// tsformat property
				if ( strLine.toLowerCase().startsWith( STR_TSFORMAT_PROPERTY ) )
				{
					int intEqualsIndex = strLine.indexOf( STR_EQUALS );
					_strDateFormat = strLine.substring( intEqualsIndex + 1 ).trim();
					try
					{
						_dateFormat = new SimpleDateFormat( _strDateFormat );
					}
					catch( Exception e )
					{
						System.out.println( "Debug::Debug() Incorrect timestamp format ; using standard format : " + e.getMessage() );
					}

					continue;
				}

				// class to log
				_logObjects.add( strLine );
			}

			buf.close();
		}
		catch( Exception e )
		{
			System.out.println( "Debug::Debug() Problem reading config file ; debug mode not activated" );
			e.printStackTrace();
			_blnActivated = false;
		}
	}

	////////////////////////////////////////////////////////////////////////////////////////
	//  Constructor
	////////////////////////////////////////////////////////////////////////////////////////

	// not instantiable
	private Debug()
	{
	}


	////////////////////////////////////////////////////////////////////////////////////////
	//  "Begin" static methods
	////////////////////////////////////////////////////////////////////////////////////////

	/**
	 *	Logs onto the standard output stream a trace following the pattern <i>formatted_ts full_classname::method_name() begin<i> , where : <br>
	 *    <i>formatted_ts</i> is the current timestamp formatted with the format defined in the config file by the property tsformat<br>
	 *    <i>full_classname</i> is the class name containing its package <br>
	 *    <i>method_name</i> is the method name<br>
	 *  In the example <i>2008/08/16 33:30:517 nico.util.StringFormater::main() begin</i><br>
	 *    formatted_ts = 2008/08/16 33:30:517 (format = yyyy/MM/dd mm:ss:S)<br>
	 *    full_classname = nico.util.StringFormater<br>
	 *    method_name = main<br>
	 *    @see java.text.SimpleDateFormat
	 */
	public static void begin()
	{
		if ( _blnActivated )
		{
			StackTraceElement stackTraceElement = new Exception().getStackTrace()[1];
			if ( _isClassInDebugScope( stackTraceElement.getClassName() ) )
			{
				_outStream.println( _buildCallingMethodName( stackTraceElement ) + STR_BEGIN );
			}
		}
	}

	
	/**
	 *	Logs onto the standard output stream a trace following the pattern <i>formatted_ts full_classname::method_name() begin with param_value</i> , where : <br>
	 *    <i>param_name</i> is a call to the toString() method of the <i>param</i> object 
	 *	  @see Debug.begin() for the rest of the description  
	 *    @see java.text.SimpleDateFormat
	 */
	public static void begin( Object param )
	{
		if ( _blnActivated )
		{
			StackTraceElement stackTraceElement = new Exception().getStackTrace()[1];
			if ( _isClassInDebugScope( stackTraceElement.getClassName() ) )
			{
				_outStream.println( _buildCallingMethodName( stackTraceElement ) + STR_BEGIN_WITH + param );
			}
		}
	}

	
	/**
	 *	Same description as Debug.begin( Object ) with the output of <i>param</i> as raw type instead of using toString() method
	 *	  @see Debug.begin() 
	 *	  @see Debug.begin( Object )  
	 *    @see java.text.SimpleDateFormat
	 */
	public static void begin( int param )
	{
		if ( _blnActivated )
		{
			StackTraceElement stackTraceElement = new Exception().getStackTrace()[1];
			if ( _isClassInDebugScope( stackTraceElement.getClassName() ) )
			{
				_outStream.println( _buildCallingMethodName( stackTraceElement ) + STR_BEGIN_WITH + param );
			}
		}
	}

	
	/**
	 *	Same description as Debug.begin( Object ) with the output of <i>param</i> as raw type instead of using toString() method
	 *	  @see Debug.begin() 
	 *	  @see Debug.begin( Object )  
	 *    @see java.text.SimpleDateFormat
	 */
	public static void begin( double param )
	{
		if ( _blnActivated )
		{
			StackTraceElement stackTraceElement = new Exception().getStackTrace()[1];
			if ( _isClassInDebugScope( stackTraceElement.getClassName() ) )
			{
				_outStream.println( _buildCallingMethodName( stackTraceElement ) + STR_BEGIN_WITH + param );
			}
		}
	}

	
	/**
	 *	Same description as Debug.begin( Object ) with the output of <i>param</i> as raw type instead of using toString() method
	 *	  @see Debug.begin() 
	 *	  @see Debug.begin( Object )  
	 *    @see java.text.SimpleDateFormat
	 */
	public static void begin( long param )
	{
		if ( _blnActivated )
		{
			StackTraceElement stackTraceElement = new Exception().getStackTrace()[1];
			if ( _isClassInDebugScope( stackTraceElement.getClassName() ) )
			{
				_outStream.println( _buildCallingMethodName( stackTraceElement ) + STR_BEGIN_WITH + param );
			}
		}
	}

	
	/**
	 *	Same description as Debug.begin( Object ) with the output of <i>param</i> as raw type instead of using toString() method
	 *	  @see Debug.begin() 
	 *	  @see Debug.begin( Object )  
	 *    @see java.text.SimpleDateFormat
	 */
	public static void begin( float param )
	{
		if ( _blnActivated )
		{
			StackTraceElement stackTraceElement = new Exception().getStackTrace()[1];
			if ( _isClassInDebugScope( stackTraceElement.getClassName() ) )
			{
				_outStream.println( _buildCallingMethodName( stackTraceElement ) + STR_BEGIN_WITH + param );
			}
		}
	}

	
	/**
	 *	Same description as Debug.begin( Object ) with the output of <i>param</i> as raw type instead of using toString() method
	 *	  @see Debug.begin() 
	 *	  @see Debug.begin( Object )  
	 *    @see java.text.SimpleDateFormat
	 */
	public static void begin( char param )
	{
		if ( _blnActivated )
		{
			StackTraceElement stackTraceElement = new Exception().getStackTrace()[1];
			if ( _isClassInDebugScope( stackTraceElement.getClassName() ) )
			{
				_outStream.println( _buildCallingMethodName( stackTraceElement ) + STR_BEGIN_WITH + param );
			}
		}
	}

	
	/**
	 *	Same description as Debug.begin( Object ) with the output of <i>param</i> as raw type instead of using toString() method
	 *	  @see Debug.begin() 
	 *	  @see Debug.begin( Object )  
	 *    @see java.text.SimpleDateFormat
	 */
	public static void begin( byte param )
	{
		if ( _blnActivated )
		{
			StackTraceElement stackTraceElement = new Exception().getStackTrace()[1];
			if ( _isClassInDebugScope( stackTraceElement.getClassName() ) )
			{
				_outStream.println( _buildCallingMethodName( stackTraceElement ) + STR_BEGIN_WITH + param );
			}
		}
	}


	////////////////////////////////////////////////////////////////////////////////////////
	//  "End" methods
	////////////////////////////////////////////////////////////////////////////////////////

	
	/**
	 *	Logs onto the standard output stream a trace following the pattern <i>formatted_ts full_classname::method_name() End<i> , where : <br>
	 *    <i>formatted_ts</i> is the current timestamp formatted with the format defined in the config file by the property tsformat<br>
	 *    <i>full_classname</i> is the class name containing its package <br>
	 *    <i>method_name</i> is the method name<br>
	 *  In the example <i>2008/08/16 33:30:517 nico.util.StringFormater::main() end</i><br>
	 *    formatted_ts = 2008/08/16 33:30:517 (format = yyyy/MM/dd mm:ss:S)<br>
	 *    full_classname = nico.util.StringFormater<br>
	 *    method_name = main<br>
	 *    @see java.text.SimpleDateFormat
	 */	
	public static void end()
	{
		if ( _blnActivated )
		{
			StackTraceElement stackTraceElement = new Exception().getStackTrace()[1];
			if ( _isClassInDebugScope( stackTraceElement.getClassName() ) )
			{
				_outStream.println( _buildCallingMethodName( stackTraceElement ) + STR_END );
			}
		}
	}

	
	/**
	 *	Logs onto the standard output stream a trace following the pattern <i>formatted_ts full_classname::method_name() end with param_value</i> , where : <br>
	 *    <i>param_name</i> is a call to the toString() method of the <i>toReturn</i> object 
	 *	  @see Debug.end() for the rest of the description  
	 *    @see java.text.SimpleDateFormat
	 */
	public static void end( Object toReturn )
	{
		if ( _blnActivated )
		{
			StackTraceElement stackTraceElement = new Exception().getStackTrace()[1];
			if ( _isClassInDebugScope( stackTraceElement.getClassName() ) )
			{
				_outStream.println( _buildCallingMethodName( stackTraceElement ) + STR_END_WITH + toReturn );
			}
		}
	}

	
	/**
	 *	Same description as Debug.end( Object ) with the output of <i>toReturn</i> as raw type instead of using toString() method
	 *	  @see Debug.end() 
	 *	  @see Debug.end( Object )  
	 *    @see java.text.SimpleDateFormat
	 */
	public static void end( int toReturn )
	{
		if ( _blnActivated )
		{
			StackTraceElement stackTraceElement = new Exception().getStackTrace()[1];
			if ( _isClassInDebugScope( stackTraceElement.getClassName() ) )
			{
				_outStream.println( _buildCallingMethodName( stackTraceElement ) + STR_END_WITH + toReturn );
			}
		}
	}

	
	/**
	 *	Same description as Debug.end( Object ) with the output of <i>toReturn</i> as raw type instead of using toString() method
	 *	  @see Debug.end() 
	 *	  @see Debug.end( Object )  
	 *    @see java.text.SimpleDateFormat
	 */
	public static void end( double toReturn )
	{
		if ( _blnActivated )
		{
			StackTraceElement stackTraceElement = new Exception().getStackTrace()[1];
			if ( _isClassInDebugScope( stackTraceElement.getClassName() ) )
			{
				_outStream.println( _buildCallingMethodName( stackTraceElement ) + STR_END_WITH + toReturn );
			}
		}
	}

	
	/**
	 *	Same description as Debug.end( Object ) with the output of <i>toReturn</i> as raw type instead of using toString() method
	 *	  @see Debug.end() 
	 *	  @see Debug.end( Object )  
	 *    @see java.text.SimpleDateFormat
	 */
	public static void end( long toReturn )
	{
		if ( _blnActivated )
		{
			StackTraceElement stackTraceElement = new Exception().getStackTrace()[1];
			if ( _isClassInDebugScope( stackTraceElement.getClassName() ) )
			{
				_outStream.println( _buildCallingMethodName( stackTraceElement ) + STR_END_WITH + toReturn );
			}
		}
	}

	
	/**
	 *	Same description as Debug.end( Object ) with the output of <i>toReturn</i> as raw type instead of using toString() method
	 *	  @see Debug.end() 
	 *	  @see Debug.end( Object )  
	 *    @see java.text.SimpleDateFormat
	 */
	public static void end( float toReturn )
	{
		if ( _blnActivated )
		{
			StackTraceElement stackTraceElement = new Exception().getStackTrace()[1];
			if ( _isClassInDebugScope( stackTraceElement.getClassName() ) )
			{
				_outStream.println( _buildCallingMethodName( stackTraceElement ) + STR_END_WITH + toReturn );
			}
		}
	}

	
	/**
	 *	Same description as Debug.end( Object ) with the output of <i>toReturn</i> as raw type instead of using toString() method
	 *	  @see Debug.end() 
	 *	  @see Debug.end( Object )  
	 *    @see java.text.SimpleDateFormat
	 */
	public static void end( char toReturn )
	{
		if ( _blnActivated )
		{
			StackTraceElement stackTraceElement = new Exception().getStackTrace()[1];
			if ( _isClassInDebugScope( stackTraceElement.getClassName() ) )
			{
				_outStream.println( _buildCallingMethodName( stackTraceElement ) + STR_END_WITH + toReturn );
			}
		}
	}

	
	/**
	 *	Same description as Debug.end( Object ) with the output of <i>toReturn</i> as raw type instead of using toString() method
	 *	  @see Debug.end() 
	 *	  @see Debug.end( Object )  
	 *    @see java.text.SimpleDateFormat
	 */
	public static void end( byte toReturn )
	{
		if ( _blnActivated )
		{
			StackTraceElement stackTraceElement = new Exception().getStackTrace()[1];
			if ( _isClassInDebugScope( stackTraceElement.getClassName() ) )
			{
				_outStream.println( _buildCallingMethodName( stackTraceElement ) + STR_END_WITH + toReturn );
			}
		}
	}


	////////////////////////////////////////////////////////////////////////////////////////
	//  Logging methods
	////////////////////////////////////////////////////////////////////////////////////////

	/**
	 *	Logs onto the standard output stream a trace following the pattern <i>formatted_ts full_classname::method_name() message<i> , where : <br>
	 *    <i>formatted_ts</i> is the current timestamp formatted with the format defined in the config file by the property tsformat<br>
	 *    <i>full_classname</i> is the class name containing its package <br>
	 *    <i>method_name</i> is the method name<br>
	 *    <i>message</i> is the content of the <i>strLog</i> parameter
	 *  In the example <i>2008/08/16 33:30:517 nico.util.StringFormater::main() coucou</i><br>
	 *    formatted_ts = 2008/08/16 33:30:517 (format = yyyy/MM/dd mm:ss:S)<br>
	 *    full_classname = nico.util.StringFormater<br>
	 *    method_name = main<br>
	 *    message = coucou<br>
	 *    @see java.text.SimpleDateFormat
	 */	
	public static void log( String strLog )
	{
		if ( _blnActivated )
		{
			StackTraceElement stackTraceElement = new Exception().getStackTrace()[1];
			if ( _isClassInDebugScope( stackTraceElement.getClassName() ) )
			{
				_outStream.println( _buildCallingMethodName( stackTraceElement ) + " " + strLog );
			}
		}
	}

	////////////////////////////////////////////////////////////////////////////////////////
	//  Stream definition method
	////////////////////////////////////////////////////////////////////////////////////////

	/**
	 * Sets the stream to which output the messages
	 * @param	ps	Output stream
	 */
	public static void setOutputStreaem( PrintStream ps )
	{
			_outStream = ps;
	}

	////////////////////////////////////////////////////////////////////////////////////////
	//  Private static methods
	////////////////////////////////////////////////////////////////////////////////////////

	
	/**
	 *	Returns calling method name contained in the specified Stack Trace Element 
	 *	@param	ste	A stack trace element
	 *	@return	Calling method name contained in ste
	 **/
	private static String _buildCallingMethodName( StackTraceElement ste )
	{
		String strMethodName = new String();

		if ( ste.getMethodName().equals( STR_CONSTRUCTOR ) )
		{
			int intLastDotIndex = ste.getClassName().lastIndexOf( STR_DOT);
			strMethodName += ste.getClassName().substring( intLastDotIndex + 1 );
		}
		else
		{
			strMethodName += ste.getMethodName();
		}

		return ( _dateFormat.format( new Date() ) + " " + ste.getClassName() + STR_RESOLUTION + strMethodName + STR_DOUBLE_BRACKETS );
	}

	
	/**
	 *	Tells whether the calling class is in the scope of the Debug class    
	 *	@param	strCallingClass	String representing calling class
	 *	@return	true if the calling class is in the scope of the Debug class
	 **/
	private static boolean _isClassInDebugScope( String strCallingClass )
	{
		StringTokenizer sttkPackages = new StringTokenizer( strCallingClass, STR_DOT );
		String strParentPackage = new String();
		while ( sttkPackages.hasMoreTokens() )
		{
			strParentPackage += sttkPackages.nextToken();
			if ( _logObjects.contains( strParentPackage ) )
			{
				return ( true );
			}
			strParentPackage += STR_DOT;
		}

		return ( false );
	}
}